c     This file contains support functions for the "poor man's 
c     parallel" version of Molscat version 14
c     Please report bugs to George McBane, mcbaneg@gvsu.edu

c     This version (pmpdyn.f) contains code for "dynamic dispatch",
c     where the jobs are handed out one by one to idle processes
c     until all the work is done.

c     Initial dynamic version 19 Feb 2005 G. C. McBane
c     Please report bugs to George McBane, mcbaneg@gvsu.edu

c     Version including channel count and sort on number
c     of channels, 26 Feb 2005

c     There is only rudimentary handling of failed processes;
c     a dying process must explicitly signal by calling 
c     the subroutine DIE).

c  --------------------------------------------------------------------


c     Obtain the number of processes and the ID of this process
c     with MPI calls.  In this dynamic version, leave the MPI
c     communications open to be used later.

      subroutine getid(numproc, myrank)
      implicit none
      integer numproc, myrank

      include 'mpif.h' 
      integer ierr


      call MPI_INIT( ierr ) 
      if (ierr.ne.0) then
         write(*, 100) ierr
         stop
 100     format(1x, 'Failed to start MPI: ierr = ', I3)
      end if
      
      call MPI_COMM_RANK( MPI_COMM_WORLD, myrank, ierr ) 
      if (ierr.ne.0) then
         write(*, 110) ierr
         stop
 110     format(1x, 'Failed to obtain rank: ierr = ', I3)
      end if


      call MPI_COMM_SIZE( MPI_COMM_WORLD, numproc, ierr ) 
      if (ierr.ne.0) then
         write(*, 120) ierr
         stop
 120     format(1x, 'Failed to obtain size: ierr = ', I3)
      end if

      return
      end



      subroutine dispatch_init (JTOTL, JTOTU, JSTEP, MXPAR, MSET, MHI, 
     1    JLEV, numproc, myrank, numj)

      implicit none
      include 'flagdefs.fi'

c     in molscat, next six params read in from &INPUT
      integer jtotu, jtotl, jstep, mxpar, mset, mhi, jlev(*)

      integer numproc,  myrank, numj

      write(6,600) myrank
 600  format(1x, 'Process ',I3, ' checking in.')

      if (myrank.eq.pmp_hostrank) then   !we're the dispatcher
         call dispatch(JTOTL, JTOTU, JSTEP, MXPAR, MSET, MHI, 
     1    jlev, numproc)
          stop  ! normal termination of dispatcher
      else     ! we're just another compute process; give maximum
               ! number of possible processes as upper loop limit
         numj = (jtotu-jtotl+1)*mxpar
         return
      end if
      end

c     Core dispatcher routine.  Communicates with get_task, 
c     which is called by each compute process to obtain a new JTOT/M
c     pair to work on.

      subroutine  dispatch(JTOTL, JTOTU, JSTEP, MXPAR, MSET, MHI, 
     1    jlev, numproc)

      implicit none

c     in molscat, next six params read in from &INPUT
      integer jtotu, jtotl, jstep, mxpar, mset, mhi, jlev(*)
      integer numproc  

      include 'flagdefs.fi'

      integer numcomp, numdone
      integer nextjob, currjob

      double precision time, oldtime
      integer loclistsize
      parameter(loclistsize = 20000)
      integer fulljlist(2,loclistsize), pointer(loclistsize)

      integer  j, m, i, totaljobs, count, rank, idum

      numcomp = numproc-1  !number of computational processes
      numdone = 0

      call gclock(oldtime)

      write(6,500) numcomp
 500  format(1x, 'Dispatcher: managing ',I3, ' processes.')

c     Generate list of all jobs.
      totaljobs = 0
      i = 1
      do j = jtotl, jtotu, jstep
         do m = 1, mxpar
            IF( (MSET.GT.0.AND.(M.LT.MSET.OR.M.GT.MHI)) .or. 
     1        (totaljobs .eq. loclistsize)   ) cycle
            fulljlist(1,i) = j
            fulljlist(2,i) = m
            i = i+1
            totaljobs = totaljobs+1
         end do
         if (totaljobs .eq. loclistsize) then
            write(*,*) 'Not enough storage available for full job list.'
            write(*,*) 'Truncating job list and continuing.'
            exit
         end if          
      end do

c     Sort job list according to number of channels

      call sort_jlist(fulljlist, totaljobs, jlev, pointer)

c     The main dispatch loop begins here.  We wait for
c     a process to ask for work.  Starting at the high-J
c     (that is, long job) end of the list, we hand out the jobs 
c     one at a time until we're out.  Then we start sending out pmp_done
c     flags until we've told all the computational processes we're
c     finished; then close MPI and return.

c     We have a simple mechanism for detecting prematurely 
c     deceased compute processes: if the process calls get_task
c     with a 'dying' flag, (or calls DIE, which then calls get_task), 
c     then get_request will return with
c     rank.lt.0.  We then know that one process has died.

c     If a process dies without sending a 'dying' flag, then we
c     are still in danger of deadlock.  That possibility has not
c     been dealt with.

      nextjob = totaljobs

      do while (numdone.lt.numcomp)
         call get_request(rank)
         if (rank.lt.0) then
            write(6,590) -rank
 590        format(1x,'Dispatcher: received dying signal from process ',
     1           I4,'.')    
            numdone = numdone+1
            cycle
         end if
         if (nextjob.ge.1) then
            currjob = pointer(nextjob)
c     nextjob in following lines needs to be replaced by pointer(nextjob)
            write(6,600) fulljlist(1,currjob),
     1           fulljlist(2,currjob), rank
 600        format(1x,'Dispatcher: sending job (j,m) = (',I4,',',I4,
     1           ') to process ', I4, '.')
            call send_task(rank, fulljlist(1,currjob),pmp_continue)
            nextjob = nextjob-1
         else
            write(6,610) rank
 610        format(1x,'Dispatcher: sending pmp_done to process ',I4,'.')
            call send_task(rank, idum, pmp_done)
            numdone = numdone+1
         end if
      end do

      call close_comms
      call gclock(time)
      write(6,650) time-oldtime
 650  format(1x,'Dispatcher signing off: ',F8.2,' CPU seconds used.')

      return
      end

c     Routine to receive job request.
      subroutine get_request(rank)
      implicit none

      integer rank

      include 'mpif.h'
      include 'flagdefs.fi'

      integer status(MPI_STATUS_SIZE)
      integer buffer(2), count, datatype, tag, ierr

      count = 2

      call MPI_RECV(buffer, count, MPI_INTEGER, MPI_ANY_SOURCE, 
     1     MPI_ANY_TAG, MPI_COMM_WORLD, status, ierr)

      tag = status(MPI_TAG)
      rank = status(MPI_SOURCE)
      if (tag.ne.pmp_healthy) then  !dying process
         write(6,500) tag, rank
 500     format(1x, 'get_request: received tag ', I4, ' from process ',
     1        I2)
         rank = -rank
      end if
      return
      end
         
         
c     Routine to send out job specification
      subroutine send_task(rank, jm, flag) 
   
      implicit none

      integer rank, jm(2), flag

      integer ierr

      include 'mpif.h'

      call MPI_SEND(jm,2,MPI_INTEGER, rank, flag, MPI_COMM_WORLD,
     1     ierr)
      return
      end

c     Main communication routine called from computational processes.
c     Must obtain task from dispatcher.  This function is part of the computational
c     process, not part of the dispatcher.

      subroutine get_task(myrank, jm, flag)
      implicit none

      include 'mpif.h'
      include 'flagdefs.fi'

      integer myrank, jm(2), flag

      integer status(MPI_STATUS_SIZE)
      integer buffer(2), count, tag, ierr, type

C     On entry, jm contains the last JTOT/M pair this process completed;
c     it will not be used for anything, but there's no need to mess with it.
c     The myrank input is irrelevant for the dynamic version; MPI
c     will set the source correctly in MPI_SEND.

c     Let dispatcher know we're ready for work.
      call MPI_SEND(jm, 2, MPI_INTEGER, pmp_hostrank, flag,
     1     MPI_COMM_WORLD, ierr)

      if (flag.eq.pmp_dying) then !calling process is bailing out
         call close_comms  !shut down MPI for this process
         return
      end if

c     Get task description from dispatcher.
      count = 2
      call MPI_RECV(buffer, count, MPI_INTEGER, pmp_hostrank, 
     1     MPI_ANY_TAG, MPI_COMM_WORLD, status, ierr)

      flag = status(MPI_TAG)
      if (flag.eq.pmp_done) then  !shut down MPI for this process
         call close_comms
      else !pass task back to caller
         jm(1) = buffer(1)
         jm(2) = buffer(2)
      end if
      return
      end

c     Shut down MPI.  
      subroutine close_comms
      implicit none
      
      include 'mpif.h'

      integer ierr

      call MPI_FINALIZE( ierr )
      if (ierr.ne.0) then
         write(*, 130) ierr
 130     format(1x, 'Failed to finalize MPI: ierr = ', I3)
      end if

      return
      end
      
*  Convert process ID (rank) to character *4 suffix for filenames
*     copied from Ian Bush's labDA function
******************************************************************************
******************************************************************************
*
      character*4 function suffix( number )
*
*  Converts the integer NUMBER to a four character string, left justified.
*
      integer number
*
*  Use internal file to acheive this.
*
      write( suffix, '(i4.4)' ) number
      return
*
      end
      
c     Routine to signal dispatcher that process is dying unexpectedly
      subroutine die
      implicit none
      include 'flagdefs.fi'
      
      integer jm, myrank
      
      call get_task(myrank, jm, pmp_dying)
      stop

      end

c     Sort_jlist returns an array pointer whose elements are
c     such that jlist(*, pointer(i)) gives jobs with increasing
c     numbers of channels if i goes from 1 to totaljobs.  jlev
c     is the Molscat JLEV array, needed by BASE for channel counting.
c     Everything else it needs is stored in COMMON.

      subroutine sort_jlist(jlist, totaljobs, jlev, pointer)
      implicit none
      integer totaljobs
      integer jlist(2,totaljobs), jlev(*), pointer(totaljobs)

      integer loclistsize
      parameter(loclistsize = 2000)
      
      
      integer  n(loclistsize)
      double precision rdum
      integer  idum, i

      if (totaljobs.gt.loclistsize) then
         write(6,*) 'loclistsize insufficient in sort_jlist.'
         write(6,*) 'Skipping sort and continuing.'
         return
      end if

      do i = 1, totaljobs
         pointer(i) = i

c     This messy call to BASE just obtains the number of channels
c     for each JTOT/M pair; all the *dum variables are unused by
c     BASE when LCNT is .true.  This call is duplicated within the
c     JTOT/M loop of each computational process but it's cheap
c     so the overhead should be tiny compared with the gains
c     in load balance efficiency for large jobs.

         CALL BASE (jlist(1,i),JLEV,n(i),rdum,rdum,rdum,rdum,rdum,rdum,
     1        rdum,idum,idum,idum,
     2   rdum,rdum,idum,rdum,rdum,jlist(2,i),.TRUE.,rdum,idum,0)
      end do

      call isort(n, pointer, totaljobs, 2)

      return

      end

 
c     ISORT modified to remove calls to SLATEC error library
c     and replace with terminal failures; otherwise unchanged.
c     G. C. McBane

      SUBROUTINE ISORT (IX, IY, N, KFLAG)
C***BEGIN PROLOGUE  ISORT
C***PURPOSE  Sort an array and optionally make the same interchanges in
C            an auxiliary array.  The array may be sorted in increasing
C            or decreasing order.  A slightly modified QUICKSORT
C            algorithm is used.
C***LIBRARY   SLATEC
C***CATEGORY  N6A2A
C***TYPE      INTEGER (SSORT-S, DSORT-D, ISORT-I)
C***KEYWORDS  SINGLETON QUICKSORT, SORT, SORTING
C***AUTHOR  Jones, R. E., (SNLA)
C           Kahaner, D. K., (NBS)
C           Wisniewski, J. A., (SNLA)
C***DESCRIPTION
C
C   ISORT sorts array IX and optionally makes the same interchanges in
C   array IY.  The array IX may be sorted in increasing order or
C   decreasing order.  A slightly modified quicksort algorithm is used.
C
C   Description of Parameters
C      IX - integer array of values to be sorted
C      IY - integer array to be (optionally) carried along
C      N  - number of values in integer array IX to be sorted
C      KFLAG - control parameter
C            =  2  means sort IX in increasing order and carry IY along.
C            =  1  means sort IX in increasing order (ignoring IY)
C            = -1  means sort IX in decreasing order (ignoring IY)
C            = -2  means sort IX in decreasing order and carry IY along.
C
C***REFERENCES  R. C. Singleton, Algorithm 347, An efficient algorithm
C                 for sorting with minimal storage, Communications of
C                 the ACM, 12, 3 (1969), pp. 185-187.
C***ROUTINES CALLED  XERMSG
C***REVISION HISTORY  (YYMMDD)
C   761118  DATE WRITTEN
C   810801  Modified by David K. Kahaner.
C   890531  Changed all specific intrinsics to generic.  (WRB)
C   890831  Modified array declarations.  (WRB)
C   891009  Removed unreferenced statement labels.  (WRB)
C   891009  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   900315  CALLs to XERROR changed to CALLs to XERMSG.  (THJ)
C   901012  Declared all variables; changed X,Y to IX,IY. (M. McClain)
C   920501  Reformatted the REFERENCES section.  (DWL, WRB)
C   920519  Clarified error messages.  (DWL)
C   920801  Declarations section rebuilt and code restructured to use
C           IF-THEN-ELSE-ENDIF.  (RWC, WRB)
C***END PROLOGUE  ISORT
C     .. Scalar Arguments ..
      INTEGER KFLAG, N
C     .. Array Arguments ..
      INTEGER IX(*), IY(*)
C     .. Local Scalars ..
      REAL R
      INTEGER I, IJ, J, K, KK, L, M, NN, T, TT, TTY, TY
C     .. Local Arrays ..
      INTEGER IL(21), IU(21)
C     .. External Subroutines ..
c      EXTERNAL XERMSG
C     .. Intrinsic Functions ..
      INTRINSIC ABS, INT
C***FIRST EXECUTABLE STATEMENT  ISORT
      NN = N
      IF (NN .LT. 1) THEN
         write(*,*) 'The number of values to be sorted is not positive.'
         stop
      ENDIF
C
      KK = ABS(KFLAG)
      IF (KK.NE.1 .AND. KK.NE.2) THEN
         write(*,*) 'The sort control parameter, K, is invalid.'
         stop
      ENDIF
C
C     Alter array IX to get decreasing order if needed
C
      IF (KFLAG .LE. -1) THEN
         DO 10 I=1,NN
            IX(I) = -IX(I)
   10    CONTINUE
      ENDIF
C
      IF (KK .EQ. 2) GO TO 100
C
C     Sort IX only
C
      M = 1
      I = 1
      J = NN
      R = 0.375E0
C
   20 IF (I .EQ. J) GO TO 60
      IF (R .LE. 0.5898437E0) THEN
         R = R+3.90625E-2
      ELSE
         R = R-0.21875E0
      ENDIF
C
   30 K = I
C
C     Select a central element of the array and save it in location T
C
      IJ = I + INT((J-I)*R)
      T = IX(IJ)
C
C     If first element of array is greater than T, interchange with T
C
      IF (IX(I) .GT. T) THEN
         IX(IJ) = IX(I)
         IX(I) = T
         T = IX(IJ)
      ENDIF
      L = J
C
C     If last element of array is less than than T, interchange with T
C
      IF (IX(J) .LT. T) THEN
         IX(IJ) = IX(J)
         IX(J) = T
         T = IX(IJ)
C
C        If first element of array is greater than T, interchange with T
C
         IF (IX(I) .GT. T) THEN
            IX(IJ) = IX(I)
            IX(I) = T
            T = IX(IJ)
         ENDIF
      ENDIF
C
C     Find an element in the second half of the array which is smaller
C     than T
C
   40 L = L-1
      IF (IX(L) .GT. T) GO TO 40
C
C     Find an element in the first half of the array which is greater
C     than T
C
   50 K = K+1
      IF (IX(K) .LT. T) GO TO 50
C
C     Interchange these elements
C
      IF (K .LE. L) THEN
         TT = IX(L)
         IX(L) = IX(K)
         IX(K) = TT
         GO TO 40
      ENDIF
C
C     Save upper and lower subscripts of the array yet to be sorted
C
      IF (L-I .GT. J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 70
C
C     Begin again on another portion of the unsorted array
C
   60 M = M-1
      IF (M .EQ. 0) GO TO 190
      I = IL(M)
      J = IU(M)
C
   70 IF (J-I .GE. 1) GO TO 30
      IF (I .EQ. 1) GO TO 20
      I = I-1
C
   80 I = I+1
      IF (I .EQ. J) GO TO 60
      T = IX(I+1)
      IF (IX(I) .LE. T) GO TO 80
      K = I
C
   90 IX(K+1) = IX(K)
      K = K-1
      IF (T .LT. IX(K)) GO TO 90
      IX(K+1) = T
      GO TO 80
C
C     Sort IX and carry IY along
C
  100 M = 1
      I = 1
      J = NN
      R = 0.375E0
C
  110 IF (I .EQ. J) GO TO 150
      IF (R .LE. 0.5898437E0) THEN
         R = R+3.90625E-2
      ELSE
         R = R-0.21875E0
      ENDIF
C
  120 K = I
C
C     Select a central element of the array and save it in location T
C
      IJ = I + INT((J-I)*R)
      T = IX(IJ)
      TY = IY(IJ)
C
C     If first element of array is greater than T, interchange with T
C
      IF (IX(I) .GT. T) THEN
         IX(IJ) = IX(I)
         IX(I) = T
         T = IX(IJ)
         IY(IJ) = IY(I)
         IY(I) = TY
         TY = IY(IJ)
      ENDIF
      L = J
C
C     If last element of array is less than T, interchange with T
C
      IF (IX(J) .LT. T) THEN
         IX(IJ) = IX(J)
         IX(J) = T
         T = IX(IJ)
         IY(IJ) = IY(J)
         IY(J) = TY
         TY = IY(IJ)
C
C        If first element of array is greater than T, interchange with T
C
         IF (IX(I) .GT. T) THEN
            IX(IJ) = IX(I)
            IX(I) = T
            T = IX(IJ)
            IY(IJ) = IY(I)
            IY(I) = TY
            TY = IY(IJ)
         ENDIF
      ENDIF
C
C     Find an element in the second half of the array which is smaller
C     than T
C
  130 L = L-1
      IF (IX(L) .GT. T) GO TO 130
C
C     Find an element in the first half of the array which is greater
C     than T
C
  140 K = K+1
      IF (IX(K) .LT. T) GO TO 140
C
C     Interchange these elements
C
      IF (K .LE. L) THEN
         TT = IX(L)
         IX(L) = IX(K)
         IX(K) = TT
         TTY = IY(L)
         IY(L) = IY(K)
         IY(K) = TTY
         GO TO 130
      ENDIF
C
C     Save upper and lower subscripts of the array yet to be sorted
C
      IF (L-I .GT. J-K) THEN
         IL(M) = I
         IU(M) = L
         I = K
         M = M+1
      ELSE
         IL(M) = K
         IU(M) = J
         J = L
         M = M+1
      ENDIF
      GO TO 160
C
C     Begin again on another portion of the unsorted array
C
  150 M = M-1
      IF (M .EQ. 0) GO TO 190
      I = IL(M)
      J = IU(M)
C
  160 IF (J-I .GE. 1) GO TO 120
      IF (I .EQ. 1) GO TO 110
      I = I-1
C
  170 I = I+1
      IF (I .EQ. J) GO TO 150
      T = IX(I+1)
      TY = IY(I+1)
      IF (IX(I) .LE. T) GO TO 170
      K = I
C
  180 IX(K+1) = IX(K)
      IY(K+1) = IY(K)
      K = K-1
      IF (T .LT. IX(K)) GO TO 180
      IX(K+1) = T
      IY(K+1) = TY
      GO TO 170
C
C     Clean up
C
  190 IF (KFLAG .LE. -1) THEN
         DO 200 I=1,NN
            IX(I) = -IX(I)
  200    CONTINUE
      ENDIF
      RETURN
      END
     

      
      
      

      
      
