c     This file contains support functions for the "poor man's 
c     parallel" version of Molscat version 14
c     Please report bugs to George McBane, mcbaneg@gvsu.edu

c     This version (pmpstat.f) contains code for "static dispatch",
c     where all the jobs are identified and allocated at
c     the beginning of the run.  

c     Obtain number of processes and the ID of this process
c     with MPI calls.  This is the only code in the static 
c     version of PMP Molscat that interacts with MPI.

      subroutine getid(numproc, myrank)
      implicit none
      integer numproc, myrank

      include 'mpif.h' 
      integer ierr


      call MPI_INIT( ierr ) 
      if (ierr.ne.0) then
         write(*, 100) ierr
         stop
 100     format(1x, 'Failed to start MPI: ierr = ', I3)
      end if
      
      call MPI_COMM_RANK( MPI_COMM_WORLD, myrank, ierr ) 
      if (ierr.ne.0) then
         write(*, 110) ierr
         stop
 110     format(1x, 'Failed to obtain rank: ierr = ', I3)
      end if


      call MPI_COMM_SIZE( MPI_COMM_WORLD, numproc, ierr ) 
      if (ierr.ne.0) then
         write(*, 120) ierr
         stop
 120     format(1x, 'Failed to obtain size: ierr = ', I3)
      end if

      call MPI_FINALIZE( ierr )
      if (ierr.ne.0) then
         write(*, 130) ierr
         stop
 130     format(1x, 'Failed to finalize MPI: ierr = ', I3)
      end if

      return
      end

      subroutine dispatch_init (JTOTL, JTOTU, JSTEP, MXPAR, MSET, MHI, 
     1    jlev, numproc, myrank, numj)

      implicit none

c     in molscat, next six params read in from &INPUT
      integer jtotu, jtotl, jstep, mxpar, mset, mhi, jlev(*)

c     At present, static version does not use jlev; it is
c     needed if we want to call BASE to count numbers of channels
c     and sort job list in that order.  It would be reasonable
c     to do that, but since I think the static version will 
c     be used mostly for CS calculations, where the sort is
c     irrelevant, I have not implemented it.

      integer numproc,  myrank, numj
      include 'tasklist.fi'  ! common block for storage of task list
c     myjlist and mynumtasks are in common; numj returned to calling routine

c     generate task list (myjlist, mynumtasks) and store in common myjlist()

      call jdist(jtotl, jtotu, jstep, mxpar, mset, mhi, 
     1     numproc, myrank, myjlist, maxnumj, numj)

      mynumtasks = numj
      return
      end

c     JTOT and M loops combined into one in pmp version

      
c     Subroutine to figure out list of JTOT and M to be done by this
c     process in a several process job. 

c     Begun 17 Jan 2005 G. C. McBane
c     Modified to handle both JTOT and M 1 Feb 2005
c     This is "static" allocation: all tasks are assigned at
c     the beginning of the run.  


      subroutine jdist(jtotl, jtotu, jstep, mxpar, mset, mhi, 
     1     numproc, myrank, myjlist, maxnumj, numj)
      implicit none

c     in molscat, next three params read in from &INPUT
      integer jtotu, jtotl, jstep, mxpar, mset, mhi

      integer  maxnumj  !size of myj array
      integer myjlist(2, maxnumj), numj, myrank, numproc

c     numproc and myrank obtained from call to getid(),
c     which obtains them from MPI calls

      integer loclistsize
      parameter(loclistsize = 20000)
      integer fulljlist(2,loclistsize)

      integer  j, m, i, pointer, totaljobs, count


c     Generate list of all jobs.
      totaljobs = 0
      i = 1
      do j = jtotl, jtotu, jstep
         do m = 1, mxpar
            IF( (MSET.GT.0.AND.(M.LT.MSET.OR.M.GT.MHI)) .or. 
     1        (totaljobs .eq. loclistsize)   ) cycle
            fulljlist(1,i) = j
            fulljlist(2,i) = m
            i = i+1
            totaljobs = totaljobs+1
         end do
         if (totaljobs .eq. loclistsize) then
            write(*,*) 'Not enough storage available for full job list.'
            write(*,*) 'Truncating job list and continuing.'
            exit
         end if          
      end do
            

c     Try to give some long and some short jobs to each process.
c     We scan through all the available jobs, selecting every
c     numproc'th one to be done by this process.  Generally the
c     jobs get longer with increasing J.
c     The (i-1) in the test statement ensures that process 0
c     will get the first job, consistent with earlier versions.

      numj = 0
      do i = 1, totaljobs
         if (mod((i-1), numproc) .eq. myrank) then !we get this one
            numj = numj+1
            if (numj .gt. maxnumj) then
               write(*,*) 'Too many jobs; truncating list.'
               exit
            end if
            myjlist(1,numj) = fulljlist(1,i)
            myjlist(2,numj) = fulljlist(2,i)
         end if
      end do
               
c     end of calculation; say hello and print results

      write(6, 100) myrank, numproc, 
     1     (myjlist(1,i), myjlist(2,i), i = 1, numj)    
 100  format(1x, 'Process ', I3, ' of ', I3, ': (JTOT, M) pairs ',/, 
     1     8('(',I3,',',I3,')') )
      
      return
      end

c     Supply (j,m) pair to process when asked.  In static
c     version this is nearly trivial.

      subroutine get_task(rank, jm, flag)
      implicit none
      integer rank, flag
      integer jm(2)
      
      include 'tasklist.fi'
c     myjlist and mynumtasks are in common
      include 'flagdefs.fi'
      
      integer nextj
      data nextj /1/
      save nextj
      
      if (nextj .gt. mynumtasks) then
         flag = pmp_done
      else
         jm(1) = myjlist(1,nextj)
         jm(2) = myjlist(2,nextj)
         nextj = nextj+1
         flag = pmp_continue
      end if
      return
      end


*  Convert process ID (rank) to character *4 suffix for filenames
*     copied from Ian Bush's labDA function
******************************************************************************
******************************************************************************
*
      character*4 function suffix( number )
*
*  Converts the integer NUMBER to a four character string, left justified.
*
      integer number
*
*  Use internal file to acheive this.
*
      write( suffix, '(i4.4)' ) number
      return
*
      end

c     die() for static version: use simple stop
      subroutine die
      stop
      end
